from enum import Enum


class EnsembleClassifierMLModel(Enum):
    LOGISTIC_REG = "logistic_reg"
    SVM = "svm"
    XGBOOST = "xgboost"

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"'{self.value}'"


class EnsembleMethod(Enum):
    GREEDY = "greedy"
    AVERAGING = "averaging"
    WEIGHTED_AVERAGING = "weighted_averaging"
    MAJORITY_VOTING = "majority_voting"
    WEIGHTED_MAJORITY_VOTING = "weighted_majority_voting"

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"'{self.value}'"

class Transformations(Enum):
    LOG = 'log'
    SQUARE = 'square'
    CUBIC = 'cubic'
    SQUARE_ROOT = 'square_root'
    CUBIC_ROOT = 'cubic_root'
    SIGMOID = 'sigmoid'
    TANH = 'tanh'

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"'{self.value}'"  # Represent the value in dict keys, etc.



import random

import numpy as np
import pandas as pd
from deap import base, creator, tools

from feature_gen.implementation.constants import EnsembleMethod
from feature_gen.implementation.ensemble_classifier import EnsembleClassifier


def run_nsga_2_algorithm_for_dataset(
        X_train: pd.DataFrame,
        X_validation: pd.DataFrame,
        y_train: pd.DataFrame,
        y_validation: pd.DataFrame,
        ensemble_method: EnsembleMethod,
        **kwargs
):
    display_each_individual = kwargs.get("display_each_individual", True)
    generations_num = kwargs.get("generations_num", 5)
    first_population_size = kwargs.get("first_population_size", 10)

    def eval_features(individual):
        if display_each_individual:
            print(individual)
        # individual is like [0, 1, 0, 1, 0] so features_vector will be like [False, True, False, True, False]
        features_vector = [bool(i) for i in individual]
        assert len(features_vector) == len(individual)
        if features_vector.count(True) == 0:
            return 0, len(individual)  # Penalize having no features selected
        ensemble_for_individual = EnsembleClassifier(
            X_train=X_train.loc[:, features_vector],
            X_test=X_validation.loc[:, features_vector],
            y_train=y_train,
            y_test=y_validation,
            **kwargs
        )
        ensemble_method_score, all_ensemble_ml_models_scores = ensemble_for_individual.run_for_individual(ensemble_method)
        return ensemble_method_score, features_vector.count(True)

    # Check if 'FitnessMulti' and 'Individual' have already been created
    if not hasattr(creator, "FitnessMulti"):
        creator.create("FitnessMulti", base.Fitness, weights=(1.0, -1.0))

    if not hasattr(creator, "Individual"):
        creator.create("Individual", list, fitness=creator.FitnessMulti)

    toolbox = base.Toolbox()
    toolbox.register("attr_bool", random.randint, 0, 1)
    toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_bool, n=X_train.shape[1])
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("evaluate", eval_features)
    toolbox.register("mate", tools.cxUniform, indpb=0.5)
    toolbox.register("mutate", tools.mutFlipBit, indpb=0.02)
    toolbox.register("select", tools.selNSGA2)

    def inner_function():
        pop = toolbox.population(n=first_population_size)
        hof = tools.ParetoFront()
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", np.mean, axis=0)
        stats.register("std", np.std, axis=0)
        stats.register("min", np.min, axis=0)
        stats.register("max", np.max, axis=0)

        # Evaluate the initial population
        fitnesses = map(toolbox.evaluate, pop)
        for ind, fit in zip(pop, fitnesses):
            ind.fitness.values = fit

        # Genetic algorithm loop
        for gen in range(1, generations_num + 1):
            offspring = []
            while len(offspring) < len(pop):
                parent1, parent2 = random.sample(pop, 2)
                if random.random() < 1.0:  # Crossover probability is 1.0
                    child1, child2 = toolbox.clone(parent1), toolbox.clone(parent2)
                    toolbox.mate(child1, child2)
                    del child1.fitness.values
                    del child2.fitness.values
                    offspring.extend([child1, child2])
                if len(offspring) > len(pop):
                    offspring = offspring[:len(pop)]

            # Apply mutation
            for mutant in offspring:
                if random.random() < 0.02:
                    toolbox.mutate(mutant)
                    del mutant.fitness.values

            # Evaluate the offspring with invalid fitness
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            for ind in invalid_ind:
                ind.fitness.values = toolbox.evaluate(ind)

            # Combine the population with offspring and select the best to form the new population
            pop[:] = toolbox.select(pop + offspring, len(pop))

            hof.update(pop)
            record = stats.compile(pop)
            if gen % 10 == 0:
                print(f"Generation {gen}: Stats {record}")

        return pop, stats, hof

    pop, stats, hof = inner_function()
    return pop, stats, hof



from typing import List

import numpy as np
import pandas as pd

from feature_gen.implementation.msga_2.nsga2_algorithm import run_nsga_2_algorithm_for_dataset


def _create_bootstrap_samples(
        X_train: pd.DataFrame,
        y_train: pd.DataFrame,
        bootstrap_samples_count: int,
        random_state: int
):
    # Initialize lists to store bootstrap samples and out-of-bag (OOB) samples.
    samples = []
    samples_oob = []
    rng = np.random.RandomState(random_state)  # Initialize RNG only once

    for _ in range(bootstrap_samples_count):
        total_samples = X_train.shape[0]

        # Generate random indices with replacement to create a bootstrap sample
        indices = rng.choice(total_samples, total_samples, replace=True)

        # Identify the out-of-bag (OOB) indices (those not selected in the bootstrap sample)
        indices_out_of_bag = np.setdiff1d(np.arange(total_samples), indices)

        # Extract the bootstrap sample for both features (X) and labels (y)
        X_sample = X_train.iloc[indices]
        y_sample = y_train.iloc[indices]

        # Extract the out-of-bag sample for both features (X) and labels (y)
        X_oob_sample = X_train.iloc[indices_out_of_bag]
        y_oob_sample = y_train.iloc[indices_out_of_bag]

        # Append the bootstrap and OOB samples to their respective lists
        samples.append((X_sample, y_sample))
        samples_oob.append((X_oob_sample, y_oob_sample))

    # Return the lists of bootstrap samples and OOB samples
    # TODO: samples are for training and samples_oob for validation?
    return samples, samples_oob


def _macro_process(feature_sets):
    # Combine all the feature sets from the micro process -> union
    optimal_features_in = np.max(feature_sets, axis=0)

    return optimal_features_in


# Feature Set Aggregation: Combines feature subsets obtained from different runs of the genetic algorithm using majority voting.
def get_the_most_optimal_features_set(X_train, y_train, ensemble_method, **kwargs) -> List[bool]:
    bootstrap_samples_count = kwargs.get('bootstrap_samples_count', 2)
    random_state = kwargs.get('random_state', 42)

    bootstrap_samples, bootstrap_oob = _create_bootstrap_samples(X_train, y_train, bootstrap_samples_count,
                                                                 random_state)
    best_feature_sets = []

    for sample_i in range(bootstrap_samples_count):
        new_x_train, new_y_train = bootstrap_samples[sample_i]
        X_validation, y_validation = bootstrap_oob[sample_i]
        pop, stats, hof = run_nsga_2_algorithm_for_dataset(
            X_train=new_x_train,
            X_validation=X_validation,
            y_train=new_y_train,
            y_validation=y_validation,
            ensemble_method=ensemble_method,
            **kwargs
        )
        best_feature_sets.append((hof[0], stats))

    # Combine all best individuals from the micro stage. Here is our Fz
    combined_features = np.array([ind for ind, _ in best_feature_sets])

    combined_features = _macro_process(combined_features)

    features_vector = [bool(i) for i in combined_features]
    new_x_train, new_y_train = bootstrap_samples[0]
    X_validation, y_validation = bootstrap_oob[0]

    # Handle unexpected classes inside EnsembleClassifier
    pop, stats, hof = run_nsga_2_algorithm_for_dataset(
        X_train=new_x_train,
        X_validation=X_validation,
        y_train=new_y_train,
        y_validation=y_validation,
        ensemble_method=ensemble_method,
        **kwargs
    )
    new_optimal_features_vector = [bool(i) for i in hof[0]]
    # The length of new_optimal_features_vector may have decreased so we use
    # Asserting the features vector length has remained unchanged
    assert len(new_optimal_features_vector) == len(features_vector) == new_x_train.shape[1] == X_validation.shape[1]
    return new_optimal_features_vector



import concurrent.futures
import random
import warnings

import numpy as np
import pandas as pd
from sklearn.exceptions import DataConversionWarning, ConvergenceWarning
from sklearn.kernel_approximation import RBFSampler
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.metrics import f1_score
from sklearn.pipeline import Pipeline
from xgboost import XGBClassifier

from feature_gen.implementation.constants import EnsembleClassifierMLModel, EnsembleMethod

# Suppress only DataConversionWarning
warnings.filterwarnings("ignore", category=DataConversionWarning)
warnings.filterwarnings("ignore", category=ConvergenceWarning)


class EnsembleClassifier:
    """
    X_train here is after the reduction of the feature space . So this class is not responsible for ignoring the features
    """

    def __init__(self, X_train: pd.DataFrame, X_test: pd.DataFrame, y_train: pd.DataFrame, y_test: pd.DataFrame,
                 **kwargs):
        self._X_train = X_train
        self._X_test = X_test
        self._y_train = y_train
        self._y_test = y_test

        # Setting up defaults and overwriting with kwargs if provided
        random_state = kwargs.get('random_state', 42)
        max_iter = kwargs.get('max_iter', 500)
        C = kwargs.get('C', 1e5)
        solver = kwargs.get('solver', 'liblinear')
        gamma = kwargs.get('gamma', 1)
        n_components = kwargs.get('n_components', 100)
        sgd_loss = kwargs.get('sgd_loss', 'hinge')
        sgd_max_iter = kwargs.get('sgd_max_iter', 1000)
        sgd_tol = kwargs.get('sgd_tol', 1e-2)
        xgb_n_estimators = kwargs.get('xgb_n_estimators', 100)

        # Set the seed for Python's built-in random module
        random.seed(random_state)
        # Set the seed for NumPy
        np.random.seed(random_state)

        # Initialize models with either provided or default parameters
        self._logistic_reg = LogisticRegression(random_state=random_state, max_iter=max_iter, C=C, solver=solver)
        self._svm = Pipeline(steps=[
            ("rbf_feature", RBFSampler(gamma=gamma, random_state=random_state, n_components=n_components)),
            ("sgd_clf", SGDClassifier(loss=sgd_loss, max_iter=sgd_max_iter, tol=sgd_tol))
        ])
        self._xgb = XGBClassifier(random_state=random_state, n_estimators=xgb_n_estimators)

        self._result = {}

    def run_for_individual(self, ensemble_method: EnsembleMethod):
        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            args = (self._X_train, self._y_train)
            res = {
                EnsembleClassifierMLModel.LOGISTIC_REG: executor.submit(self._logistic_reg.fit, *args),
                EnsembleClassifierMLModel.SVM: executor.submit(self._svm.fit, *args),
                EnsembleClassifierMLModel.XGBOOST: executor.submit(self._xgb.fit, *args)
            }

            for classifier, future in res.items():
                res[classifier] = future.result()

            predictions_dict = {}
            # Collect predictions from each classifier
            for classifier, model in res.items():
                predictions_dict[classifier] = model.predict(self._X_test)

        # 1. Greedy (Grd)
        if ensemble_method == EnsembleMethod.GREEDY:
            best_classifier = max(
                res.keys(),
                key=lambda clf: f1_score(self._y_test, predictions_dict[clf], average='macro')
            )
            self._result[EnsembleMethod.GREEDY] = predictions_dict[best_classifier]

        # 2. Averaging (Avg)
        if ensemble_method == EnsembleMethod.AVERAGING:
            avg_pred = np.mean(
                [predictions_dict[clf] for clf in res.keys()], axis=0
            )
            self._result[EnsembleMethod.AVERAGING] = np.round(avg_pred).astype(int)

        if ensemble_method == EnsembleMethod.WEIGHTED_AVERAGING:
            # 3. Weighted Averaging (Wavg)
            weights = {  # Weights based on accuracy
                clf: f1_score(self._y_test, predictions_dict[clf], average='macro')
                for clf in res.keys()
            }
            wavg_pred = np.average(
                [predictions_dict[clf] for clf in res.keys()], axis=0, weights=[weights[clf] for clf in res.keys()]
            )
            self._result[EnsembleMethod.WEIGHTED_AVERAGING] = np.round(wavg_pred).astype(int)

        # 4. Majority Voting
        if ensemble_method == EnsembleMethod.MAJORITY_VOTING:
            unique_classes = np.unique(self._y_train)
            majority_votes = np.zeros([len(res.keys()), self._X_test.shape[0], len(unique_classes)])
            idx = 0
            for clf in res.keys():
                for idxLabel in range(len(unique_classes)):
                    majority_votes[idx, predictions_dict[clf] == unique_classes[idxLabel], idxLabel] += 1
                idx += 1
            self._result[EnsembleMethod.MAJORITY_VOTING] = unique_classes[
                np.argmax(np.sum(majority_votes, axis=0), axis=1)
            ]

        # 5. Weighted Majority Voting (Wmv)
        if ensemble_method == EnsembleMethod.WEIGHTED_MAJORITY_VOTING:
            weights = {  # Weights based on accuracy
                clf: f1_score(self._y_test, predictions_dict[clf], average='macro')
                for clf in res.keys()
            }
            weighted_majority_votes = []
            for i in range(len(self._X_test)):
                class_votes = {}
                for clf in res.keys():
                    vote = predictions_dict[clf][i]
                    if vote not in class_votes:
                        class_votes[vote] = 0
                    class_votes[vote] += weights[clf]

                weighted_majority_votes.append(
                    max(class_votes, key=class_votes.get)
                )
            self._result[EnsembleMethod.WEIGHTED_MAJORITY_VOTING] = np.array(weighted_majority_votes)

        predicted_for_ensemble_method = self._result[ensemble_method]
        ensemble_method_score = f1_score(self._y_test, predicted_for_ensemble_method, average='macro')
        # Now based on all the ensemble methods we find the accuracy OR f1-score. We decided to go with f1-score
        all_ensemble_ml_models_scores = {
            cls: f1_score(self._y_test, predictions_dict[cls], average='macro') for cls, pred_set in
            predictions_dict.items()
        }
        return ensemble_method_score, all_ensemble_ml_models_scores



import datetime
from typing import List

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder

from feature_gen.implementation.constants import EnsembleMethod
from feature_gen.implementation.ensemble_classifier import EnsembleClassifier
from feature_gen.implementation.get_the_most_optimal_features_set import get_the_most_optimal_features_set
from feature_gen.utils.checker import check_if_this_is_classification_problem
from feature_gen.utils.df_preprocessor import DFPreprocessor


class FeatureGenMaster:

    def __init__(self, df: pd.DataFrame, target_column_name: str, min_number_of_target_unique_values: int = 20):
        self.df = df
        self._y = df[target_column_name]
        self._X = df.drop(columns=[target_column_name], axis=1)

        self._apply_one_hot_encoding_on_target_columns()

        if not check_if_this_is_classification_problem(
                y=self._y,
                min_number_of_target_unique_values=min_number_of_target_unique_values
        ):
            raise ValueError("This problem is not a classification problem.")

        self._all_best_features = {}
        self._best_original_features = {}
        self._best_new_features = {}

        self._average_all_ensemble_methods_scores = {}
        self._all_ensemble_methods_scores = {}

    def _apply_one_hot_encoding_on_target_columns(self) -> None:
        # This iis needed for Xgboost
        label_encoder = LabelEncoder()
        y = label_encoder.fit_transform(self._y)
        self._y = pd.DataFrame({self._y.name: y})

    def _pre_process_df(self) -> pd.DataFrame:
        df_preprocessor = DFPreprocessor(X=self._X)
        X = (
            df_preprocessor.
            drop_unnamed().
            drop_infinity().
            one_hot_encode().
            fill_nulls_with_mean().
            remove_potential_stratification().
            standardize_data().
            normalize_data().
            drop_highly_correlated_features().
            apply_transformations().
            get_final_X()
        )
        return X

    @staticmethod
    def _get_real_names_of_features(
            most_optimal_features_set: List[bool],
            pre_processed_and_transformed_X: pd.DataFrame
    ) -> List[str]:
        return [
            feature_name for feature_name, feature in
            zip(pre_processed_and_transformed_X.columns, most_optimal_features_set) if feature
        ]

    def _start(self, ensemble_methods: List[EnsembleMethod], **kwargs) -> None:
        test_size = kwargs.get('test_size', 0.2)
        random_state = kwargs.get('random_state', 42)

        dataset_original_features_count_before_ts = self._X.shape[1]
        pre_processed_and_transformed_X = self._pre_process_df()
        dataset_original_features_count_after_ts = pre_processed_and_transformed_X.shape[1]

        for ensemble_method in ensemble_methods:
            print(f"Ensemble method {ensemble_method} is being considered")
            date0 = datetime.datetime.now()
            X_train, X_test, y_train, y_test = train_test_split(
                pre_processed_and_transformed_X,
                self._y,
                test_size=test_size,
                random_state=random_state,
                stratify=self._y
            )
            most_optimal_features_set = get_the_most_optimal_features_set(
                X_train=X_train,
                y_train=y_train,
                ensemble_method=ensemble_method,
                **kwargs
            )

            new_X_train_for_most_optimal_features_set = X_train.iloc[:, most_optimal_features_set]
            new_X_test_for_most_optimal_features_set = X_test.iloc[:, most_optimal_features_set]

            ensemble_accuracies = EnsembleClassifier(
                X_train=new_X_train_for_most_optimal_features_set,
                X_test=new_X_test_for_most_optimal_features_set,
                y_train=y_train,
                y_test=y_test,
                **kwargs
            )
            ensemble_method_score, all_ensemble_ml_models_scores = ensemble_accuracies.run_for_individual(
                ensemble_method)
            r = {
                "dataset_original_features_count_before_ts": dataset_original_features_count_before_ts,
                "dataset_original_features_count_after_ts": dataset_original_features_count_after_ts,
                "dataset_features_count_after_whole_process": new_X_train_for_most_optimal_features_set.shape[1],
                "execution_time": (datetime.datetime.now() - date0).seconds,
                "ensemble_method_score": ensemble_method_score,
                "all_ensemble_ml_models_scores": all_ensemble_ml_models_scores
            }
            self._all_best_features[ensemble_method] = self._get_real_names_of_features(
                most_optimal_features_set=most_optimal_features_set,
                pre_processed_and_transformed_X=pre_processed_and_transformed_X,
            )
            self._all_ensemble_methods_scores[ensemble_method] = ensemble_method_score

        for ensemble_method in ensemble_methods:
            self._best_original_features[ensemble_method] = [
                f for f in self._all_best_features[ensemble_method] if f in self._X.columns
            ]
            self._best_new_features[ensemble_method] = [
                f for f in self._all_best_features[ensemble_method] if f not in self._X.columns
            ]

    def start(self, ensemble_methods: List[EnsembleMethod], **kwargs) -> None:
        self._start(ensemble_methods, **kwargs)

    def get_all_best_features(self):
        if not self._all_best_features:
            raise ValueError("You should call start() function first")
        return self._all_best_features

    def get_best_original_features(self):
        if not self._best_original_features:
            raise ValueError("You should call start() function first")
        return self._best_original_features

    def get_best_new_features(self):
        if not self._best_new_features:
            raise ValueError("You should call start() function first")
        return self._best_new_features

    def get_all_ensemble_methods_scores(self):
        if not self._all_ensemble_methods_scores:
            raise ValueError("You should call start() function first")
        return self._all_ensemble_methods_scores



from setuptools import setup, find_packages

setup(
    name="feature_gen",
    version="0.1.3",
    author="UWinnipeg2025",
    # author_email="armin.felahat.cs@gmail.com",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    # url="https://github.com/ArminCS97/masters_project",
    packages=find_packages(),
    install_requires=[
        "deap==1.3.2",
        "pandas==2.2.3",
        "scikit-learn==1.5.2",
        "xgboost==2.1.2"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



import pandas as pd
import numpy as np


def check_if_this_is_classification_problem(
        y: pd.DataFrame,
        min_number_of_target_unique_values: int = 20
) -> bool:
    """
    Determines if a given problem is a classification problem based on the number of unique target values.

    Args:
        y (pd.DataFrame): The target values in a DataFrame or Series format.
        min_number_of_target_unique_values (int): The minimum number of unique target values to consider the problem
                                                  as regression rather than classification.

    Returns:
        bool: True if the problem is a classification problem, False otherwise.
    """
    if not isinstance(y, (pd.Series, pd.DataFrame, np.ndarray)):
        raise ValueError("The target 'y' must be a pandas Series, DataFrame, or numpy array.")

    # Convert DataFrame or 2D array to 1D array
    if isinstance(y, pd.DataFrame):
        if y.shape[1] != 1:
            raise ValueError("The target 'y' DataFrame must contain exactly one column.")
        y = y.squeeze()  # Convert DataFrame to Series

    # Ensure y is 1D
    if len(y.shape) > 1:
        y = y.ravel()  # Flatten to 1D

    # Count unique values in the target
    num_unique_values = pd.Series(y).nunique()

    # Classification is determined if unique values are fewer than the threshold
    return num_unique_values < min_number_of_target_unique_values



from typing import Self

import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler, StandardScaler


class DFPreprocessor:
    def __init__(self, X: pd.DataFrame) -> None:
        self._X = X.copy()  # Store a copy to avoid modifying the original data
        self.__update_column_types()

    def __update_column_types(self):
        self._numeric_columns = self._X.select_dtypes(include='number').columns
        self._non_numeric_columns = self._X.select_dtypes(exclude='number').columns

    def drop_unnamed(self):
        if 'Unnamed: 0' in self._X.columns:
            self._X = self._X.drop(columns=['Unnamed: 0'])
        self.__update_column_types()
        return self

    def one_hot_encode(self) -> Self:
        """Converts all boolean and string columns into numeric format using one-hot encoding."""
        if isinstance(self._X, pd.DataFrame):
            # Apply one-hot encoding
            self._X = pd.get_dummies(self._X, columns=self._non_numeric_columns, drop_first=False)
        elif isinstance(self._X, np.ndarray):
            raise ValueError("For one-hot encoding, convert self._X to a DataFrame.")
        else:
            raise TypeError("self._X must be a pandas DataFrame for this operation.")
        return self

    def drop_infinity(self) -> Self:
        """Drop the rows with infinity."""
        self._X = self._X[~((self._X == np.inf) | (self._X == -np.inf)).any(axis=1)]
        self.__update_column_types()
        return self

    def fill_nulls_with_mean(self) -> Self:
        """Make the null values filled with the mean of the respective columns."""
        self._X = self._X.fillna(self._X.mean())
        return self

    # TODO: Maybe the question is a binary alssifcication
    def remove_potential_stratification(self) -> Self:
        """Removes potential stratification in the dataset."""
        # Assuming stratification means ensuring that the target variable is balanced or that there's no unintended bias
        # You can adjust the logic based on your specific definition of stratification
        # This is just an example of dropping any rows where the target column is the same as others.
        # Replace 'target' with the actual name of your target column if needed

        # if self._X[self._target_column_name].nunique() <= 2:
        #     self._X = self._X.drop(columns=[self._target_column_name])
        return self

    def standardize_data(self) -> Self:
        """Standardizes numerical columns to have a mean of 0 and standard deviation of 1."""
        # If we have no numeric columns, we simply skip over this step to avoid "ValueError: at least one array or dtype is required"
        if len(self._numeric_columns) == 0:
            return self
        scaler = StandardScaler()
        self._X[self._numeric_columns] = scaler.fit_transform(self._X[self._numeric_columns])
        return self

    def normalize_data(self) -> Self:
        """Normalizes numerical columns to be in the range [0, 1]."""
        # If we have no numeric columns, we simply skip over this step to avoid "ValueError: at least one array or dtype is required"
        if len(self._numeric_columns) == 0:
            return self
        scaler = MinMaxScaler()
        self._X[self._numeric_columns] = scaler.fit_transform(self._X[self._numeric_columns])
        return self

    def drop_highly_correlated_features(self, threshold=0.9) -> Self:
        """Removes columns with correlation greater than the threshold."""
        corr_matrix = self._X.corr().abs()
        upper = corr_matrix.where(np.triu(np.ones(corr_matrix.shape), k=1).astype(bool))
        to_drop = [column for column in upper.columns if any(upper[column] > threshold)]
        self._X = self._X.drop(columns=to_drop)
        self.__update_column_types()
        return self

    def apply_transformations(self) -> Self:
        if len(self._numeric_columns) == 0:
            return self

        epsilon = 1e-10
        transformed_cols = []

        # Attention. the transformations are applied only on originally numeric values
        for col in self._numeric_columns:
            # Calculate all transformations and store them in temporary DataFrames
            col_log = np.log(self._X[col] + np.abs(np.min(self._X[col])) + epsilon)
            col_squared = self._X[col] ** 2
            col_cubic = self._X[col] ** 3
            col_sqrt = np.sqrt(self._X[col] + np.abs(np.min(self._X[col])))
            col_cbrt = np.cbrt(self._X[col])

            # Sigmoid transformation (maps values to range [0, 1])
            z = (self._X[col] - self._X[col].min()) / (self._X[col].max() - self._X[col].min() + epsilon)
            z = 2 * z - 1
            col_sigmoid = 1 / (1 + np.exp(-z))

            # Tanh transformation (maps values to range [-1, 1])
            col_tanh = np.tanh(z)

            # Store each transformation in a list with appropriate column names
            transformed_cols.append(pd.DataFrame({
                f"{col}_log": col_log,
                f"{col}_squared": col_squared,
                f"{col}_cubic": col_cubic,
                f"{col}_sqrt": col_sqrt,
                f"{col}_cbrt": col_cbrt,
                f"{col}_sigmoid": col_sigmoid,
                f"{col}_tanh": col_tanh,
            }))

        # Concatenate all the transformed columns to the original DataFrame
        transformed_df = pd.concat(transformed_cols, axis=1)
        self._X = pd.concat([self._X, transformed_df], axis=1).copy()

        return self

    def get_final_X(self) -> pd.DataFrame:
        return self._X
